from .helpers import Utils


class Migration():

    def __init__(self, version, extension, name, checksum):
        self.version = version
        self.extension = extension
        self.name = name
        self.checksum = checksum

    @classmethod
    def from_name(cls, name, **kwargs):
        version = kwargs.get('version', Utils.get_version_from_name(name))
        extension = kwargs.get('extension', Utils.get_extension_from_name(name))
        checksum = kwargs.get('checksum', Utils.load_checksum_from_name(name))
        return cls(version, extension, name, checksum)

    @classmethod
    def from_list(cls, list_):
        return [cls(m.version, m.extension, m.name, m.checksum) for m in list_]

    @classmethod
    def from_list_names(cls, list_):
        return [cls.from_name(m.name) for m in list_]

    def __repr__(self):
        return 'Migration(version=%s, extension=%s, name=%s, checksum=%s)' % \
            (self.version, self.extension, self.name, self.checksum)



from .log import logger
from .helpers import Utils
from .validate import Validate
from .migration import Migration
from .dbms.database import factory
from .errors import MIGRATIONS_NOT_FOUND


class Migrate():

    def __init__(self, conf):
        Validate(conf).run()
        self._db = factory(conf.DBMS)(conf)
        self._migration_dir = conf.DATABASE_MIGRATION_DIR

    def run(self):
        migrations_to_be_executed = self._get_migration_files_to_be_executed()
        if not migrations_to_be_executed:
            logger.info("Nothing to do")
            return

        for migration in migrations_to_be_executed:
            logger.info("Migrating --> %s" % migration.name)
            try:
                with open(Utils.fullname(migration.name), "r") as sqlfile:
                    self._db.execute(sqlfile.read())
                self._db.upgrade_version(migration)
                logger.success("%s SUCCESS" % migration.name)
            except Exception as error:
                logger.error(error)

    def _get_migration_files_to_be_executed(self):
        all_local_migrations = self._get_all_local_migrations()
        all_db_migrations = Migration.from_list(self._db.get_all_schema_migrations())
        if all_db_migrations and not all_local_migrations:
            logger.error(MIGRATIONS_NOT_FOUND % self._migration_dir)
        return Utils.subtract(all_local_migrations, all_db_migrations)

    def _get_all_local_migrations(self):
        local_files = Utils.get_local_files()
        if not local_files:
            return []
        migrations = [Migration.from_name(local_file) for local_file in local_files]
        return Utils.sort_migrations_list(migrations)



import os

from .log import logger
from .settings import DATABASE_MIGRATION_DIR, LOGS_DIR


def _dir_maker(dir_):
    if not os.path.exists(dir_):
        logger.info("%s don't found, creating local folder" % dir_)
        os.makedirs(dir_)


_dir_maker(LOGS_DIR)
_dir_maker(DATABASE_MIGRATION_DIR)

'''
import os
import sys
import logging
import dotenv

logger = logging.getLogger(__name__)

dotenv.load_dotenv()
ENVIRONMENT = os.getenv('ENVIRONMENT')

if not ENVIRONMENT:
    logger.error('Required ENVIRONMENT env variable')
    sys.exit(1)

env_dotfile = os.path.join(os.getcwd(), '.env.' + ENVIRONMENT)

if not os.path.exists(env_dotfile):
    logger.warn('%s not found. Environment variables will be used.' % env_dotfile)
else:
    logger.info('Loading dotenv for environment %s: %s', ENVIRONMENT, env_dotfile)
    dotenv.load_dotenv(dotenv_path=env_dotfile)
'''



from .log import logger
from .helpers import Utils
from .dbms.database import factory
from .migration import Migration
from .errors import OUT_OF_DATE_ERROR, DIFF_NAME_ERROR, DIFF_CHECKSUM_ERROR, VALID_NAME_ERROR, MIGRATIONS_NOT_FOUND


class Validate():

    def __init__(self, conf):
        self._db = factory(conf.DBMS)(conf)
        self._migration_dir = conf.DATABASE_MIGRATION_DIR

    def run(self):
        local_migrations = self._get_all_local_migrations()
        db_migrations = Migration.from_list(self._db.get_all_schema_migrations())

        if db_migrations and not local_migrations:
            logger.error(MIGRATIONS_NOT_FOUND % self._migration_dir)

        if local_migrations:
            local_migrations_map = Utils.create_map_from_list("version", local_migrations)

            for db_migration in db_migrations:
                logger.info("Validating --> %s" % db_migration.name)
                local_migration = local_migrations_map.get(db_migration.version)
                self._out_of_date(local_migration, db_migration)
                self._name_format(local_migration.name)
                self._diff_names(local_migration, db_migration)
                self._diff_checksum(local_migration, db_migration)
                logger.success("%s VALID" % db_migration.name)

    def _out_of_date(self, local_migration, db_migration):
        if local_migration is None:
            logger.error(OUT_OF_DATE_ERROR % db_migration.name)

    def _diff_names(self, local_migration, db_migration):
        if local_migration.name != db_migration.name:
            logger.error(DIFF_NAME_ERROR % (local_migration.name, db_migration.name))

    def _diff_checksum(self, local_migration, db_migration):
        if local_migration.checksum != db_migration.checksum:
            error = DIFF_CHECKSUM_ERROR % (local_migration.name, local_migration.checksum, db_migration.checksum)
            logger.error(error)

    def _name_format(self, name):
        if not Utils.is_file_name_valid(name):
            error = VALID_NAME_ERROR % (name, Utils.expected_pattern())
            logger.error(error)

    def _get_all_local_migrations(self):
        local_files = Utils.get_local_files()
        if not local_files:
            return []
        migrations = [Migration.from_name(local_file) for local_file in local_files]
        return Utils.sort_migrations_list(migrations)



import os
from .version import __version__


# INITIALIZATION
LOGO = """
     ██████╗ ██╗   ██╗██╗    ██╗ █████╗ ██╗   ██╗
     ██╔══██╗╚██╗ ██╔╝██║    ██║██╔══██╗╚██╗ ██╔╝
     ██████╔╝ ╚████╔╝ ██║ █╗ ██║███████║ ╚████╔╝
     ██╔═══╝   ╚██╔╝  ██║███╗██║██╔══██║  ╚██╔╝
     ██║        ██║   ╚███╔███╔╝██║  ██║   ██║
     ╚═╝        ╚═╝    ╚══╝╚══╝ ╚═╝  ╚═╝   ╚═╝
                        %s
""" % __version__


# PYWAY VARIABLES
DATABASE_MIGRATION_DIR = os.environ.get('PYWAY_DATABASE_MIGRATION_DIR', 'resources')
SQL_MIGRATION_PREFIX = os.environ.get('PYWAY_SQL_MIGRATION_PREFIX', 'V')
SQL_MIGRATION_SEPARATOR = os.environ.get('PYWAY_SQL_MIGRATION_SEPARATOR', '__')
SQL_MIGRATION_SUFFIXES = os.environ.get('PYWAY_SQL_MIGRATION_SUFFIXES', '.sql')
TABLE = os.environ.get('PYWAY_TABLE', 'public.schema_version')
DBMS = os.environ['PYWAY_DBMS']
LOGS_DIR = os.getenv('PYWAY_LOGS_DIR', 'logs')


# DATABASE VARIABLES
DATABASE_URL = os.environ['DATABASE_URL']
DATABASE_PORT = os.environ.get('DATABASE_PORT', '5432')
DATABASE_NAME = os.environ['DATABASE_NAME']
DATABASE_USERNAME = os.environ['DATABASE_USERNAME']
DATABASE_PASSWORD = os.environ['DATABASE_PASSWORD']
DATABASE_CONNECT_TIMEOUT = os.environ.get('DATABASE_CONNECT_TIMEOUT', '30')
DATABASE_MAX_CONNECTIONS = os.environ.get('DATABASE_MAX_CONNECTIONS', '10')



import click

from pyway import settings
from pyway.info import Info
from pyway.log import logger
from pyway.migrate import Migrate
from pyway.validate import Validate


@click.group()
def cli():
    logger.info(settings.LOGO)


@cli.command()
def migrate():
    logger.info('STARTING MIGRATE PROCESS . . .')
    Migrate(settings).run()
    logger.info('.MIGRATE ENDED.')


@cli.command()
def validate():
    logger.info('STARTING VALIDATE PROCESS . . .')
    Validate(settings).run()
    logger.info('.VALIDATE ENDED.')


@cli.command()
def info():
    logger.info('INFO . . .')
    Info(settings).run()
    logger.info('.INFO ENDED.')


if __name__ == '__main__':
    cli()



from pydbwrapper import Database as Datasource, Config


CREATE_VERSION_MIGRATIONS = "create table if not exists %s ("\
    "installed_rank serial PRIMARY KEY,"\
    "version varchar(20) NOT NULL,"\
    "extension varchar(20) NOT NULL,"\
    "name varchar(125) NOT NULL,"\
    "checksum varchar(25) NOT NULL"\
    ");"
SELECT_FIELDS = ("version", "extension", "name", "checksum")
ORDER_BY_FIELD_ASC = "installed_rank"
ORDER_BY_FIELD_DESC = "installed_rank desc"
INSERT_VERSION_MIGRATE = "insert into %s (version, extension, name, checksum) values (%s, %s, %s, %s);"


class Postgres():

    def __init__(self, config):
        self.config = config
        self.version_table = config.TABLE
        self.create_version_table_if_not_exists()

    def connect(self):
        return Datasource(Config(config_dict={
            "host": self.config.DATABASE_URL,
            "port": self.config.DATABASE_PORT,
            "dbname": self.config.DATABASE_NAME,
            "user": self.config.DATABASE_USERNAME,
            "password": self.config.DATABASE_PASSWORD,
            "connect_timeout": int(self.config.DATABASE_CONNECT_TIMEOUT),
            "maxconnections": int(self.config.DATABASE_MAX_CONNECTIONS)
        }))

    def create_version_table_if_not_exists(self):
        self.execute(CREATE_VERSION_MIGRATIONS % self.version_table)

    def execute(self, scipt):
        with self.connect() as db:
            db.execute(scipt, None, True)

    def get_all_schema_migrations(self):
        with self.connect() as db:
            return db.select(self.version_table)\
                .fields(*SELECT_FIELDS)\
                .order_by(ORDER_BY_FIELD_ASC).execute().fetchall()

    def upgrade_version(self, migration):
        with self.connect() as db:
            db.insert(self.version_table).setall(migration.__dict__).execute()

    def get_max_version(self):
        with self.connect() as db:
            fetched = db.select(self.version_table).fields("version").order_by(ORDER_BY_FIELD_DESC)\
                .execute().fetchone()
            return int(fetched.version) if fetched is not None else 0
